from queue import Empty
from typing import Any, Generator, Tuple

from .consumer import _ConsumerMixin
from .producer import _ProducerMixin
from .task import Task


class Join(Task, _ProducerMixin[Tuple[Any, ...]], _ConsumerMixin[Any]):
    """Join Task."""

    def __init__(self) -> None:
        """"""
        Task.__init__(self)
        _ProducerMixin.__init__(self)
        _ConsumerMixin.__init__(self)

    def _inlet(self) -> Generator[Tuple[Any, ...], None, None]:
        while self._is_running():
            try:
                results = []
                for in_queue in self.in_queues:
                    while self._is_running():
                        try:
                            i = in_queue.get(timeout=1)
                            results.append(i)
                            break
                        except Empty:
                            pass
                if len(self.in_queues) == len(results):
                    yield tuple(results)
                else:
                    assert not self._is_running()
            except GeneratorExit:
                break

    def run(self) -> None:
        """Run and start the activity"""
        for i in self._inlet():
            o = i
            self._outlet(o)
            if not self._is_running():
                break



import base64
import copy
import io
import os
import socket
from threading import Lock
from typing import Optional

from PIL.Image import Image as PIL_Image

from .schema.agent_app_protocol import CommandKind, CommandRequest, CommandResponse, Status
from .task import Isolated


class CommandServer(Isolated):
    sock_path: Optional[str]
    img_lock: Lock
    img: Optional[PIL_Image]

    """Actcast Command Server

    This server handles these commands

    * 'Take Photo'
        * responses cached image as png data

    """

    def __init__(self, sock_path: Optional[str] = None) -> None:
        super().__init__()
        self.sock_path = None
        env = "ACTCAST_COMMAND_SOCK"
        if env in os.environ:
            self.sock_path = os.environ[env]
        if sock_path is not None:
            self.sock_path = sock_path
        self.running = True
        self.img_lock = Lock()
        self.img = None

    def run(self) -> None:
        """Run and start the activity"""
        if self.sock_path is None:
            return
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            os.unlink(self.sock_path)
        except FileNotFoundError:
            pass  # ignore
        s.bind(self.sock_path)
        s.settimeout(1)
        s.listen(1)
        while self.running:
            # Wait photo
            while self.running:
                with self.img_lock:
                    if self.img is None:
                        continue
                    else:
                        break
            try:
                assert self.img is not None

                conn, _ = s.accept()
                request, err = CommandRequest.parse(conn)
                if err:
                    response = CommandResponse(
                        copy.copy(request.id_),
                        Status.GENERAL_ERROR,
                        b"",
                    )
                    conn.sendall(response.to_bytes())
                    continue

                if request.kind == CommandKind.TAKE_PHOTO:
                    header = b"data:image/png;base64,"
                    with self.img_lock:
                        pngimg = io.BytesIO()
                        self.img.save(pngimg, format="PNG")
                        b64img = base64.b64encode(pngimg.getbuffer())
                    data = header + b64img
                    response = CommandResponse(copy.copy(request.id_), Status.OK, data)
                else:
                    response = CommandResponse(
                        copy.copy(request.id_),
                        Status.GENERAL_ERROR,
                        b"",
                    )
                conn.sendall(response.to_bytes())
                conn.shutdown(socket.SHUT_RDWR)
                conn.close()
            except socket.timeout:
                pass
        os.remove(self.sock_path)

    def update_image(self, image: PIL_Image) -> None:
        """

        Update the cached 'Take Photo' command image.

        Args:
            image (:class:`~PIL.Image`): image

        """
        with self.img_lock:
            self.img = image.copy()



from queue import Full
from typing import Generic, List, TypeVar

from ..util.pad import _PadBase, _PadBlocking, _PadIn
from .consumer import _ConsumerMixin
from .task import Task, _TaskI

T_OUT = TypeVar("T_OUT")


class _ProducerMixin(Generic[T_OUT], _TaskI):
    out_queues: List[_PadIn[T_OUT]]
    out_queue_id: int

    def __init__(self) -> None:
        """"""
        self.out_queues = []
        self.out_queue_id = 0

    def _add_out_queue(self, q: _PadIn[T_OUT]) -> None:
        self.out_queues.append(q)

    def _new_pad(self) -> _PadBase[T_OUT]:
        return _PadBlocking()

    def connect(self, follow: _ConsumerMixin[T_OUT]) -> None:
        """
        Connect following task.
        """
        assert isinstance(follow, _ConsumerMixin)

        pad_out, pad_in = self._new_pad().into_pad_pair()
        follow._add_in_queue(pad_out)
        self._add_out_queue(pad_in)

    def _outlet(self, o: T_OUT) -> bool:
        length = len(self.out_queues)
        while self._is_running():
            try:
                self.out_queues[self.out_queue_id].put(o, timeout=1)
                self.out_queue_id = (self.out_queue_id + 1) % length
                return True
            except Full:
                pass
        return False


class Producer(Generic[T_OUT], Task, _ProducerMixin[T_OUT]):
    """Producer Task."""

    def __init__(self) -> None:
        """"""
        Task.__init__(self)
        _ProducerMixin.__init__(self)

    def run(self) -> None:
        """Run and start the activity"""
        while True:
            o = self.proc()
            self._outlet(o)
            if not self._is_running():
                break

    def proc(self) -> T_OUT:
        """
        Pipeline Task Processor
        """
        raise NotImplementedError("'proc' must be overridden.")



import json
from pathlib import Path
from typing import Any, Callable, Dict, List

from . import capture, task  # noqa: F401
from .application import Application  # noqa: F401
from .command_server import CommandServer  # noqa: F401
from .service_client import ServiceClient  # noqa: F401


def notify(
    notification: List[Dict[str, Any]],
    *args: Any,
    **kwargs: Any,
) -> None:
    """

    Make a notification to Actcast.

    Args:
        notification (list of dict): dicts must be encodable to JSON.

    Example:

        >>> import actfw_core
        >>> actfw_core.notify([{'msg': 'Hello!'}])
        [{"msg": "Hello!"}]

    """
    if type(notification) != list:
        raise TypeError("must be a list of JSON encodable objects.")
    kwargs["flush"] = True
    print(json.dumps(notification), *args, **kwargs)


_default_heartbeat_file = Path("/root/heartbeat")


def _default_heartbeat(*args: Any, **kwargs: Any) -> None:
    _default_heartbeat_file.touch()


_heartbeat_function = _default_heartbeat


def set_heartbeat_function(f: Callable[..., None]) -> None:
    """

    Set heartbeat action.

    Args:
        f (function): function which execute heartbeat action

    Example:

        >>> import actfw_core
        >>> def heartbeat(): print("working!")
        ...
        >>> actfw_core.set_heartbeat_function(heartbeat)
        >>> actfw_core.heartbeat()
        working!

    """
    global _heartbeat_function
    _heartbeat_function = f


def heartbeat(*args: Any, **kwargs: Any) -> None:
    """

    Execute heartbeat action.

    Notes:
        Default action is 'touch /root/heartbeat'.

    """
    _heartbeat_function(*args, **kwargs)



from typing import Generic, TypeVar

from .consumer import _ConsumerMixin
from .producer import _ProducerMixin
from .task import Task

T_OUT = TypeVar("T_OUT")
T_IN = TypeVar("T_IN")


class Pipe(Generic[T_OUT, T_IN], Task, _ProducerMixin[T_OUT], _ConsumerMixin[T_IN]):
    """Straightforward Pipeline Task."""

    def __init__(self) -> None:
        """"""
        Task.__init__(self)
        _ProducerMixin.__init__(self)
        _ConsumerMixin.__init__(self)

    def run(self) -> None:
        """Run and start the activity"""
        for i in self._inlet():
            o = self.proc(i)
            self._outlet(o)
            if not self._is_running():
                break

    def proc(self, i: T_IN) -> T_OUT:
        """
        Pipeline Task Processor
        """
        raise NotImplementedError("'proc' must be overridden.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['actfw_core',
 'actfw_core._private.agent_app_protocol',
 'actfw_core._private.compat',
 'actfw_core._private.schema',
 'actfw_core._private.util',
 'actfw_core.schema',
 'actfw_core.task',
 'actfw_core.util',
 'actfw_core.v4l2']

package_data = \
{'': ['*']}

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.8,<0.9'],
 ':python_version < "3.8"': ['Pillow>=5,<6'],
 ':python_version >= "3.8"': ['Pillow>=8,<9']}

setup_kwargs = {
    'name': 'actfw-core',
    'version': '2.2.0a0',
    'description': 'Core components of actfw, independent of specific devices',
    'long_description': "# actfw-core\n\nCore components of actfw, a framework for Actcast Application written in Python.\nactfw-core is intended to be independent of any specific device.\n\n## Installation\n\n```console\nsudo apt-get update\nsudo apt-get install -y python3-pip python3-pil \nsudo apt-get install -y libv4l-0 libv4lconvert0  # if using `V4LCameraCapture`\npip3 install actfw-core\n```\n\n## Document\n\n* [API References](https://idein.github.io/actfw-core/latest/)\n\n## Usage\n\nConstruct your application with a task parallel model\n\n* Application\n  * `actfw_core.Application` : Main application\n* Workers\n  * `actfw_core.task.Producer` : Task generator\n    * `actfw_core.capture.V4LCameraCapture` : Generate UVC camera capture image\n  * `actfw_core.task.Pipe` : Task to Task converter\n  * `actfw_core.task.Consumer` : Task terminator\n\nEach worker is executed in parallel.\n\nUser should\n\n* Define subclass of `Producer/Pipe/Consumer`\n\n```python\nclass MyPipe(actfw_core.task.Pipe):\n    def proc(self, i):\n        ...\n```\n\n* Connect defined worker objects\n\n```python\np  = MyProducer()\nf1 = MyPipe()\nf2 = MyPipe()\nc  = MyConsumer()\np.connect(f1)\nf1.connect(f2)\nf2.connect(c)\n```\n\n* Register to `Application`\n\n```python\napp = actfw_core.Application()\napp.register_task(p)\napp.register_task(f1)\napp.register_task(f2)\napp.register_task(c)\n```\n\n* Execute application\n\n```python\napp.run()\n```\n\n## Development Guide\n\n### Installation of dev requirements\n\n```console\npip3 install poetry\npoetry install\n```\n\n### Running tests\n\n```console\npoetry run nose2 -v\n```\n\n### Releasing package & API doc\n\nCI will automatically do.\nFollow the following branch/tag rules.\n\n1. Make changes for next version in `master` branch (via pull-requests).\n2. Make a PR that updates version in `pyproject.toml` and merge it to `master` branch.\n3. Create Git tag from `master` branch's HEAD named `release-<New version>`. E.g. `release-1.4.0`.\n4. Then CI will build/upload package to PyPI & API doc to GitHub Pages.\n",
    'author': 'Idein Inc.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Idein/actfw-core',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
